from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional

from .client import ApiClient, client
from .types import Schedule


@dataclass
class Task:
    """
    This class implements the structure of Tasks of the VC Publisher API.

    # Internal attributes
    This are attributes, which are not overgiven by the API, but are necessary
    for the object to work. They might not be initialized.

    :attribute _api: The API client
    :atype _api: ApiClient
    :attribute _endpoint: The API endpoint
    :atype _endpoint: str

    # Required attributes
    This are attributes, which are definitly overgiven by the API.

    :attribute _id: task id
    :atype _id: str
    :attribute createdAt: task creation date
    :atype createdAt: datetime
    :attribute updatedAt: task update date
    :atype updatedAt: datetime
    :attribute createdBy: task creator
    :atype createdBy: str
    :attribute updatedBy: task last updater
    :atype updatedBy: str
    :attribute labels: task labels
    :atype labels: list
    :attribute properties: task properties
    :atype properties: dict
    :attribute tags: task tags: additional metadata attributes not used by the Publisher
    :atype tags: dict
    :attribute debugLevel: task debug level
    :atype debugLevel: int
    :attribute jobType: task job type
    :atype jobType: str
    :attribute jobVersion: task job version
    :atype jobVersion: str
    :attribute projectId: project id
    :atype projectId: str
    :attribute priority: task priority
    :atype priority: int
    :attribute parameters: task parameters
    :atype parameters: dict
    :attribute schedule: task schedule
    :atype schedule: dict

    # Optional attributes, which are not necessarily overgiven by the API
    :attribute name: task name
    :atype name: str
    :attribute description: task description
    :atype description: str
    :attribute lastJobId: task last job id
    :atype lastJobId: str
    """

    # Internal attributes
    _api: ApiClient = field(default=client, init=False, repr=False)
    _endpoint: str = field(init=False, repr=False)

    # Required attributes
    _id: str
    createdAt: datetime
    updatedAt: datetime
    createdBy: str
    updatedBy: str
    labels: list[str]
    properties: dict
    tags: dict
    debugLevel: int
    jobType: str
    jobVersion: str
    projectId: str
    priority: int
    parameters: dict
    schedule: Schedule

    # Optional attributes
    name: Optional[str]
    description: Optional[str]
    lastJobId: Optional[str]

    ############## Dunder Methods ##############
    def __post_init__(self):
        """
        Initialize the API endpoint, after the object is created.
        """
        self._endpoint = f'project/{self.projectId}/task/{self._id}/'

    def __str__(self):
        """
        String representation of the Task object as its id.
        """
        return self._id



from typing import Any, Optional

from httpx import AsyncClient, Client, Response, post

from .auth import BearerAuth
from .Settings import settings
from .types import ApiClientProtocol


def log(event_name, info):
    """
    Logging function for httpx client trace extension.
    """
    print(event_name, info)


class ApiClient(ApiClientProtocol):
    _instance = None
    _connected = False
    _url: str = ''

    def __new__(cls):
        """
        Singleton Pattern
        """
        if cls._instance is None:
            cls._instance = super(ApiClient, cls).__new__(cls)
        return cls._instance

    def __login__(self) -> bool:
        """
        Login to API
        :return: bearer token
        """
        if not self._connected:
            bearer: str = 'no bearer'
            self._url: str = f'{settings.host}/api/{settings.api_version}/'
            if self._url:
                response = post(
                    url=self._url + 'login/',
                    data={
                        'username': settings.user,
                        'password': settings.password,
                    },
                )
                if response.status_code == 200:
                    bearer = response.json()['token']
                    self._client = Client(base_url=f'{self._url}/')
                    self._aclient = AsyncClient(base_url=f'{self._url}/')
                    self._client.auth = BearerAuth(bearer)
                    self._aclient.auth = BearerAuth(bearer)
                    self._connected = True
                else:
                    raise Exception(f'Login failed: {response.__dict__}')
        return self._connected

    def __logout__(self) -> None:
        """
        logout from API
        """
        if self._connected:
            response = self._client.get(url=self._url + 'logout/')
            # should return 201 Logout Successful
            if response.status_code == 201:
                # self.logger.debug("Logout.")
                print('Logout.')
            else:
                # self.logger.warning(f"Logout failed: {response.json()}")
                print(f'Logout failed: {response.__dict__}')

    def get(
        self, endpoint: str, params: Optional[dict] = None, *args, **kwargs
    ) -> Response:
        """
        Make a GET Request to the VC Publisher API.

        :param endpoint: api endpoint like `projects/`
        :param headers:
        :param stream: just for file downloads, default False
        :return: Response
        """

        def get_it():
            """
            Get Request
            """
            url = self._url + endpoint
            response: Response = self._client.get(
                url=url,
                params=params,
                extensions={},
            )
            return response

        if self._connected:
            return get_it()
        else:
            if self.__login__():
                return get_it()
            else:
                response = Response(status_code=502)
                return response

    def post(
        self,
        endpoint: str,
        data: Optional[dict] = None,
        json: Optional[dict] = None,
        params: Optional[dict] = None,
        files: Optional[Any] = None,
        *args,
        **kwargs,
    ) -> Response:
        """
        Make a POST Request to the VC Publisher API.

        :param endpoint: api endpoint like `project/`
        :param data: dictionary delivered in request body
        :param json:
        :param params:
        :param files:
        :return:
        """

        def post_it():
            """
            Post Request
            """
            url: str = self._url + endpoint
            response = self._client.post(
                url=url,
                data=data,
                json=json,
                params=params,
                files=files,
                extensions={},
            )
            return response

        if self._connected:
            return post_it()
        else:
            if self.__login__():
                return post_it()
            else:
                response = Response(status_code=502)
                return response

    def delete(
        self,
        endpoint: str,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
    ) -> Response:
        """
        Make a DELETE Request to the VC Publisher API.

        :param endpoint: api endpoint like `project/<project_id>/`
        :type endpoint: str
        :param headers: Optional dict for headers
        :type headers: Optional[dict]
        :param params: Optional dict for query parameters
        :type params: Optional[dict]
        :return: Response as dict
        :rtype: Response
        """

        def delete_it():
            """
            Delete Request
            """
            url: str = self._url + endpoint
            response = self._client.delete(
                url=url,
                headers=headers,
                params=params,
                extensions={'trace': log},
            )
            return response

        if self._connected:
            return delete_it()
        else:
            if self.__login__():
                return delete_it()
            else:
                response = Response(status_code=502)
                return response

    def put(
        self,
        endpoint: str,
        data: Optional[dict] = None,
        json: Optional[dict] = None,
        params: Optional[dict] = None,
        files: Optional[Any] = None,
    ) -> Response:
        """
        Make a PUT Request to the VC Publisher API.

        :param endpoint: The endpoint to PUT to.
        :type endpoint: str
        :param data: The data to PUT.
        :type data: Optional[dict]
        :param json: The JSON data to PUT.
        :type json: Optional[dict]
        :param params: The parameters to PUT.
        :type params: Optional[dict]
        :param files: The files to PUT.
        :type files: Optional[Any]
        :return: The response from the API.
        """

        def put_it():
            """
            Put Request
            """
            url = self._url + endpoint
            response = self._client.put(
                url=url,
                data=data,
                json=json,
                params=params,
                files=files,
                extensions={'trace': log},
            )
            return response

        if self._connected:
            return put_it()
        else:
            if self.__login__():
                return put_it()
            else:
                response = Response(status_code=502)
                return response

    async def stream(
        self,
        endpoint: str,
        params: Optional[dict] = None,
    ) -> Any:
        """
        Führt einen Streaming-Request aus und gibt einen Generator zurück,
        der über die Bytes des Response iteriert. Der HTTP-Stream wird hier
        innerhalb eines with-Blocks geöffnet und automatisch geschlossen,
        wenn der Generator erschöpft ist.
        """

        async def stream_it():
            """
            Stream Request
            """
            url = self._url + endpoint
            return self._aclient.stream(
                method='GET',
                url=url,
                params=params,
            )

        if self._connected:
            return stream_it()
        else:
            if self.__login__():
                return stream_it()
            else:
                response = Response(status_code=502)
                return response


client = ApiClient()



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class User:
    _id: str
    username: str
    email: str
    createdAt: Optional[datetime] = None
    updatedAt: Optional[datetime] = None
    createdBy: Optional[str] = ''
    updatedBy: Optional[str] = ''



from dataclasses import dataclass, field
from datetime import datetime
from typing import Literal, Optional

from .client import client
from .types import ApiClientProtocol, SourceProperty


@dataclass
class Source:
    """
    This class implements the structure of Datasources of the VC Publisher API.
    """

    # Internal attributes
    _api: ApiClientProtocol = field(default=client, init=False, repr=False)
    _endpoint: str = field(init=False, repr=False)

    # required api attributes
    _id: str
    name: str
    properties: dict
    typeProperties: dict
    sourceProperties: SourceProperty
    type: Literal[
        'tileset',
        'tilesetupdate',
        'geojson',
        'oblique',
        'qmesh',
        'meshinmesh',
        'wms',
        'wmts',
        'tms',
        'vectortiles',
        'generic',
    ]
    dataUpdatedAt: datetime
    dataUpdatedBy: str | None  # api could return it with value null
    projectId: str
    uri: str
    jobIds: list[str]
    publishTaskIds: list[str]

    # optional api attributes
    description: Optional[str] = ''
    bbox: Optional[list[float]] = None

    def publish(
        self,
        credentialsId: str,
        destination: str,
        numThreads: Optional[int],
        gzip: Optional[bool],
    ):
        """
        Not implemented yet.
        Publish the datasource with the given parameters.
        """
        pass

    ############## Dunder Methods ##############
    def __post_init__(self):
        """
        This function is called after the initialization of the object.
        """
        self._endpoint = f'project/{self.projectId}/datasource/{self._id}/'

    def __str__(self) -> str:
        return self._id



from .Bucket import Bucket as Bucket
from .core import get_project as get_project
from .Project import Project as Project
from .Settings import settings as settings
from .Source import Source as Source
from .Task import Task as Task



import os
from datetime import datetime

from tqdm import tqdm

from .types import ExternalSource, InternalSource, SourceProperty


############## Dacite Type-Hooks ##############
def parse_datetime(value: str) -> datetime:
    """
    Type-hook for dacite to parse datetime from isoformat.
    It replaces the 'Z' with '+00:00' to make it compatible with fromisoformat.

    :param value: datetime in isoformat
    :type value: str
    :return: datetime
    :rtype: datetime
    """
    return datetime.fromisoformat(value.replace('Z', '+00:00'))


def parse_source_property(value: dict) -> SourceProperty:
    """
    Type-hook for dacite to parse SourceProperty from dict.

    :param value: SourceProperty as dict
    :type value: dict
    :return: SourceProperty
    :rtype: SourceProperty
    """
    if value['type'] == 'external':
        return ExternalSource(**value)
    elif value['type'] == 'internal':
        return InternalSource(**value)
    else:
        raise ValueError(f'Unknown SourceProperty type: {value["type"]}')


############## other ##############
def file_upload_generator(filepath: str):
    """
    Generator to upload a file with progress bar.

    :yield: file-like object
    :rtype: file-like object
    """
    total = os.path.getsize(filepath)
    with tqdm(
        ascii=True, unit_scale=True, unit='B', unit_divisor=1024, total=total
    ) as bar:
        with open(filepath, 'rb') as file:
            while chunk := file.read(1024):
                bar.update(len(chunk))
                yield chunk
